% Hekate Inference Engine

% run multiple xtts, in turn; can be extended to run them in parallel
run_xtt_multiple(XttIds):-
	forall(member(XttId,XttIds),run_xtt_single(XttId)).

% run inference process at given XTT:
% 1. find conflict set,
% 2. resolve conflict set
% 3. fire rules
% 4. follow rule links
run_xtt_single(XttId):-
	(bheart(debug) -> write('run_xtt_single:'), write(XttId),nl; true),
	perform_read_io(XttId),
	find_conflict_set(XttId,Rules),
%	write('run_xtt_single: conflict set:'), write(Rules),nl,
	resolve_conflict_set(Rules,RulesToFire),
%	write('run_xtt_single: rules to fire:'), write(RulesToFire),nl,
	fire_multiple(RulesToFire),
	perform_write_io(XttId),
%	write('run_xtt_single: end'),nl,
	findall(Xtt,(member(Rule,RulesToFire),rule(_,Rule,_,_,Xtt)),Xtts),
	flatten(Xtts,XttsToFollow),
%	write('run_xtt_single: XttsToFollow:'),write(XttsToFollow),nl,
	!,
	run_xtt_multiple(XttsToFollow).


find_conflict_set(XttId,Rules):-
	findall(Rule,(rule(XttId,Rule,Conditions,_,_),check_conditions(Conditions)),Rules).

check_conditions([]).
check_conditions(Conditions):-
	forall(member([Operator|Args],Conditions),logop(Operator,Args)).

% resolve conflict set, choose, in order:
% 1. rules, which have no outgoing links,
% 2. first rule with an outgoing link
resolve_conflict_set([R|_],[R]):-
	rule(_,R,_,_,[_|_]).
resolve_conflict_set([],[]).
resolve_conflict_set([R|Rules],[R|RulesToFire]):-
	rule(_,R,_,_,[]),
	resolve_conflict_set(Rules,RulesToFire).

% fire rules with given IDs
fire_multiple(Rules):-
	forall(member(Rule,Rules),fire_single(Rule)).
fire_single(Rule):-
	(bheart(debug) -> write('firing rule:'), write(Rule),nl ; true),
	rule(_,Rule,_,Decision,_),
	forall(member([Operator|Args],Decision),modop(Operator,Args)),
	(bheart(debug) -> attdispall ; true).

%I/O facilities
perform_read_io(XttId):-
	schema(XttId,Cond,_),
	forall(member(att(Att),Cond),attread(Att)).
perform_write_io(XttId):-
	schema(XttId,_,Decision),
	forall(member(att(Att),Decision),attwrite(Att)).

% gives attribute value or domain value, helper for operators
att_or_domain_value(domain(Value),Value).
att_or_domain_value(att(Att),Value):-
	attval(Att,Value).

% gives attribute or domain or evlop value, evalop helper
att_or_domain_or_evalop_value(Arg,Value):-
	att_or_domain_value(Arg,Value).
att_or_domain_or_evalop_value([Operator|Args],Value):-
	evalop(Operator,Args,Value).


