% auxiliary predicates for runtime

% WARNING: type/domain checks work for STRING and NUMERIC base types ONLY.
:- write('Prolog Runtime: WARNING: type/domain checks work for STRING and NUMERIC base types ONLY.\n').


% is attribute ro or rw?
is_read_class(AttributeId):-
	att(AttributeId,_,_,_,_,ro).
is_read_class(AttributeId):-
	att(AttributeId,_,_,_,_,rw).
% is attribute wo or rw?
is_write_class(AttributeId):-
	att(AttributeId,_,_,_,_,rw).
is_write_class(AttributeId):-
	att(AttributeId,_,_,_,_,wo).

% set attrbut to a given value
setattval(Att,Val):-
	typecheck(Att,Val),
	retractall(attval(Att,_)),
	assert(attval(Att,Val)).

%check if Val is a valid value for give attribute, checks lenght, scale, domain.
typecheck(Att,[Val]):-
	att(Att,_,_,TypeId,single,_),
	type(TypeId,_,BaseType,Len,Scale,_,Domain),
	lenscalecheck(BaseType,Len,Scale,Val), % check if val is within defined range
	domaincheck(Domain,Val). % check if val is within domian (if defined)
% for multiple valued, Val is a list of values
typecheck(Att,Val):-
	att(Att,_,_,TypeId,multiple,_),
	type(TypeId,_,BaseType,Len,Scale,_,Domain),
	maplist(lenscalecheck(BaseType,Len,Scale),Val),
	maplist(domaincheck(Domain),Val). 

% if domain is an empty list, it means that any value is acceptable
domaincheck([],_).
domaincheck(Domain,Val):-domaincheck_nonempty(Domain,Val).

domaincheck_nonempty([range(F,T)|_],Val):-
	Val @>= F, Val @=< T.
domaincheck_nonempty([Val|_],Val).
domaincheck_nonempty([_|Rest],Val):-
	domaincheck_nonempty(Rest,Val).
	
lenscalecheck(numeric,Len,Scale,Val):-
	number(Val),
	maxnumber(Len,Scale,Max),
	Val >= - Max, Val =< Max.

lenscalecheck(string,Len,_,Val):-
	is_list(Val),
	length(Val,L),
	L =< Len.

maxnumber(Len,Scale,Max):-
	Scale =< Len,
	L is Len - Scale,
	Max is 10^L - 1 + ( 1 - 10^( - Scale ) ).
	
%debug helper displaying attribute values
attdisp(AttributeId):-attval(AttributeId,Value), write(AttributeId), write(':'),
	att(AttributeId,Name,_,_,_,_), write(Name), write('='), write(Value), nl.

attdispall:-attdisp(_),fail.
attdispall.
